# -*- coding: utf-8 -*-
# vim:fenc=utf-8

'''
  Moving average based algorithms
  -------------------------------

  :copyright (c) 2014 Xavier Bruhiere.
  :license: Apache 2.0, see LICENSE for more details.
'''

from zipline.transforms import MovingAverage
import insights.plugins.mail as mail
from intuition.api.algorithm import TradingFactory
from insights.algorithms.utils import common_middlewares


class DualMovingAverage(TradingFactory):
    '''
    doc: Buys once its short moving average crosses its long moving average
      (indicating upwards momentum) and sells its shares once the averages
      cross again (indicating downwards momentum).
    parameters:
      long_window: Long window of data to track for moving average computation.
        [default 30]
      short_window: Short window of data to track for moving average
        computation. [default 0.5*long_window]
      short_rate: Alternative to short_window parameter.
        Set short window length as a fraction of the long one. [default 0.5]
      threshold: Distance between moving averages to trigger signal [default 0]
    '''
    def initialize(self, properties):
        # Interactive, mobile, hipchat, database and commission middlewares
        for middleware in common_middlewares(properties, self.identity):
            self.use(middleware)
        report_mails = properties.get('reports')
        if report_mails:
            self.use(mail.Report(report_mails).send_briefing)

        long_window = int(properties.get('long_window', 30))
        short_window = int(properties.get('short_window', 0))
        if not short_window:
            short_window = int(round(
                properties.get('ma_rate', 0.5) * float(long_window), 2))
        self.threshold = properties.get('threshold', 0)

        self.add_transform(MovingAverage, 'short_mavg', ['price'],
                           window_length=short_window)

        self.add_transform(MovingAverage, 'long_mavg', ['price'],
                           window_length=long_window)

        # To keep track of whether we invested in the stock or not
        self.invested = {}
        self.short_mavgs = []
        self.long_mavgs = []

    def warm(self, data):
        self.invested = {sid: False for sid in data}

    def event(self, data):
        self.logger.debug('Processing event on {}'.format(self.get_datetime()))
        signals = {'buy': {}, 'sell': {}}

        for ticker in data:
            short_mavg = data[ticker].short_mavg['price']
            long_mavg = data[ticker].long_mavg['price']

            if short_mavg - long_mavg > self.threshold \
                    and not self.invested[ticker]:
                signals['buy'][ticker] = data[ticker]
                self.invested[ticker] = True

            elif short_mavg - long_mavg < -self.threshold \
                    and self.invested[ticker]:
                signals['sell'][ticker] = data[ticker]
                self.invested[ticker] = False

        return signals


# https://www.quantopian.com/posts/this-is-amazing
'''
class Momentum(TradingFactory):
    # FIXME Too much transactions, can't handle it on wide universe
    def initialize(self, properties):
        # Interactive, mobile, hipchat, database and commission middlewares
        for middleware in common_middlewares(properties, self.identity):
            self.use(middleware)

        self.max_notional = 2000.1
        self.min_notional = -2000.0

        self.max_weight = properties.get('max_weight', 0.2)
        self.max_exposure = properties.get('max_exposure', self.max_weight)

        self.add_transform(MovingAverage, 'mavg', ['price'],
                           window_length=properties.get('window', 3))

    def event(self, data):
        signals = {'buy': {}, 'sell': {}}

        for ticker in data:
            sma = data[ticker].mavg.price
            price = data[ticker].price
            cash = self.portfolio.cash
            notional = self.portfolio.positions[ticker].amount * price
            capital_used = self.portfolio.capital_used

            if sma > price and \
                    notional > -self.max_exposure * (capital_used + cash):
                signals['sell'][ticker] = data[ticker]
            elif sma < price and \
                    notional < self.max_weight * (capital_used + cash):
                signals['buy'][ticker] = data[ticker]

        return signals
'''



# -*- coding: utf-8 -*-
# vim:fenc=utf-8

'''
  Mongodb context
  ---------------

  Fetch the configuration from mongo documents.
  It needs an input like :
  <1.2.3.4:port/database/collection/doc_id>

  :copyright (c) 2014 Xavier Bruhiere.
  :license: Apache 2.0, see LICENSE for more details.
'''

import pymongo
from intuition.errors import LoadContextFailed
from intuition.api.context import ContextFactory


# Super interface: http://genghisapp.com/
class MongodbContext(ContextFactory):

    def initialize(self, storage):
        if len(storage['path']) != 3:
            raise LoadContextFailed(
                driver=__name__,
                reason='you must provide a path like <db/collection/doc_id>')

        self.database = storage['path'][0]
        self.collection = storage['path'][1]
        self.conf_id = storage['path'][2]

        self.client = pymongo.MongoClient(
            "mongodb://{}/".format(storage['uri']))

    def _connect(self):
        if self.database not in self.client.database_names():
            raise LoadContextFailed(
                driver=__name__,
                reason='database {} not found'.format(self.database))
        return self.client[self.database]

    def _grab_doc(self, db):
        if self.collection not in db.collection_names():
            raise LoadContextFailed(
                driver=__name__,
                reason='collection {} not found'.format(self.collection))
        return db[self.collection]

    def load(self):
        db = self._connect()
        self.log.info('connected to {} database'.format(db.name))

        conf_doc = self._grab_doc(db)
        self.log.info('got {} document'.format(conf_doc.name))

        context = conf_doc.find_one({'id': self.conf_id})
        if not context:
            raise LoadContextFailed(
                driver=__name__,
                reason='document {} not found'.format(self.conf_id))

        return context



# -*- coding: utf-8 -*-
# vim:fenc=utf-8

'''
  Util plugins
  ------------

  :copyright (c) 2014 Xavier Bruhiere.
  :license: Apache 2.0, see LICENSE for more details.
'''


def debug_portfolio(logger, datetime, portfolio):
    logger.debug('[{}] {}'.format(datetime, portfolio))


def debug_metrics(logger, datetime, perf_tracker):
    if perf_tracker.progress != 0.0:
        logger.debug('[{}] {}'.format(
            datetime, perf_tracker.cumulative_risk_metrics.to_dict()))



# -*- coding: utf-8 -*-
# vim:fenc=utf-8

'''
  Manager utilities
  -----------------

  :copyright (c) 2014 Xavier Bruhiere.
  :license: Apache 2.0, see LICENSE for more details.
'''

import numpy as np
import scipy.optimize


def simplex_projection(v, b=1):
    """Projection vectors to the simplex domain

    Implemented according to the paper: Efficient projections onto the
    l1-ball for learning in high dimensions, John Duchi, et al. ICML 2008.
    Implementation Time: 2011 June 17 by Bin@libin AT pmail.ntu.edu.sg
    Optimization Problem: min_{w}\| w - v \|_{2}^{2}
    s.t. sum_{i=1}^{m}=z, w_{i}\geq 0

    Input: A vector v \in R^{m}, and a scalar z > 0 (default=1)
    Output: Projection vector w

    :Example:
    >>> proj = simplex_projection([.4 ,.3, -.4, .5])
    >>> print(proj)
    np.array([ 0.33333333, 0.23333333, 0. , 0.43333333])
    >>> print(proj.sum())
    1.0

    Original matlab implementation: John Duchi (jduchi@cs.berkeley.edu)
    Python-port: Copyright 2013 by Thomas Wiecki (thomas.wiecki@gmail.com).
    """

    v = np.asarray(v)
    p = len(v)

    # Sort v into u in descending order
    v = (v > 0) * v
    u = np.sort(v)[::-1]
    sv = np.cumsum(u)

    rho = np.where(u > (sv - b) / np.arange(1, p + 1))[0][-1]
    theta = np.max([0, (sv[rho] - b) / (rho + 1)])
    w = (v - theta)
    w[w < 0] = 0
    return w


# Weights - array of asset weights (derived from market capitalizations)
# Expreturns - expected returns based on historical data
# Covars - covariance matrix of asset returns based on historical data
def assets_meanvar(daily_returns, trading_days_per_year=255):
    # Calculate expected returns
    expreturns = np.array([])
    (rows, cols) = daily_returns.shape
    for r in range(rows):
        expreturns = np.append(expreturns, np.mean(daily_returns[r]))

    # Compute covariance matrix
    covars = np.cov(daily_returns)
    # Annualize expected returns and covariances
    expreturns = (1 + expreturns) ** trading_days_per_year - 1
    covars = covars * trading_days_per_year

    return expreturns, covars


# Solve for optimal portfolio weights
def solve_weights(R, C, rf):
    n = len(R)
    # Start optimization with equal weights
    W = np.ones([n]) / n
    # Bounds for decision variables
    b_ = [(0.1, 1) for i in range(n)]
    # Constraints - weights must sum to 1
    c_ = ({'type': 'eq', 'fun': lambda W: sum(W) - 1.})
    # 'target' return is the expected return on the market portfolio
    optimized = scipy.optimize.minimize(
        fitness, W, (
            R, C, sum(R * W)), method='SLSQP', constraints=c_, bounds=b_)
    if not optimized.success:
        #NOTE Or np.zeros ?
        return np.ones([n]) / n
        #raise BaseException(optimized.message)
    return optimized.x


# Compute the expected return of the portfolio.
def compute_mean(W, R):
    return sum(R * W)


# Compute the variance of the portfolio.
def compute_var(W, C):
    return np.dot(np.dot(W, C), W)


def fitness(W, R, C, r):
    # For given level of return r, find weights which minimizes portfolio
    # variance.
    mean_1, var = compute_mean(W, R), compute_var(W, C)
    # Penalty for not meeting stated portfolio return effectively serves as
    # optimization constraint
    # Here, r is the 'target' return
    penalty = 0.1 * abs(mean_1 - r)
    return var + penalty



# -*- coding: utf-8 -*-
# vim:fenc=utf-8

'''
  :copyright (c) 2014 Xavier Bruhiere
  :license: Apache 2.0, see LICENSE for more details.
'''

from zipline.transforms.batch_transform import batch_transform


@batch_transform
def get_past_prices(data):
    return data['price']


@batch_transform
def get_past_returns(data):
    '''
    Parameters: data
        pandas.panel (major: index, minor: sids)
    '''
    returns_df = data['price'].pct_change()
    # Because of return calculation, first raw is nan
    #FIXME nan values remain anyway
    #return np.nan_to_num(returns_df.values[1:])
    return returns_df.fillna(0.0)



# -*- coding: utf-8 -*-
# vim:fenc=utf-8

import numpy as np
from intuition.api.portfolio import PortfolioFactory


def compute_weigths(daily_returns):
    try:
        # create a covariance matrix
        covariance_matrix = np.cov(daily_returns, y=None, rowvar=1,
                                   bias=0, ddof=None)
        covariance_matrix = np.matrix(covariance_matrix)

        # calculate global minimum portfolio weights
        one_vector = np.matrix(np.ones(len(daily_returns))).transpose()
        one_row = np.matrix(np.ones(len(daily_returns)))
        covariance_matrix_inv = np.linalg.inv(covariance_matrix)
        numerator = np.dot(covariance_matrix_inv, one_vector)
        denominator = np.dot(np.dot(one_row, covariance_matrix_inv),
                             one_vector)

        return numerator / denominator
    except Exception, error:
        print(error)
        return np.zeros(len(daily_returns))


# https://www.quantopian.com/posts/global-minimum-variance-portfolio?c=1
class GlobalMinimumVariance(PortfolioFactory):
    '''
    doc: Computes a suitable compromise between risks and returns on buy
      signals and gives up the stock on sell ones.

    parameters:
      partial_sale: Percentage of the stock amount to sell [default 1]
    '''
    def initialize(self, properties):
        self.partial_sell = properties.get('partial_sell', 1.0)
        self.log.info(properties)

    def optimize(self, to_buy, to_sell):

        allocations = {}

        # Simply sell
        # NOTE It will be overwritten if we had buy signals
        for sid in to_sell:
            # NOTE Cannot go short in this configuration
            allocations[sid] = - int(round(
                self.portfolio.positions[sid].amount * self.partial_sell))

        if to_buy:
            # TODO If the universe is large enough, limit `returns` to `to_buy`
            returns = self.properties.get('historical_prices')
            # NOTE if returns.isnull().any().any(): ?
            if returns is None:
                # TODO A fallback
                raise NotImplementedError('GMV manager needs history prices')

            weights = compute_weigths(returns.transpose())
            if np.isnan(weights).any() or not weights.any():
                self.log.warning('Could not compute weigths')
            else:
                for i, sid in enumerate(returns):
                    allocations[sid] = float(weights[i])

        return allocations, 0, 1



# -*- coding: utf-8 -*-
# vim:fenc=utf-8

'''
  Intuition analysis module
  -------------------------

  Provides high level building blocks to extract insights from various datasets

  :copyright (c) 2014 Xavier Bruhiere
  :license: Apache 2.0, see LICENSE for more details.
'''

import os
import shutil
import rpy2.robjects as robjects
import dna.logging

log = dna.logging.logger(__name__)


class Stocks(object):
    ''' Produce R report of stocks opportunities '''
    knitr_report = '~/.intuition/assets/report.rnw'

    def __init__(self, report_template=None):
        log.info('loading R context')
        self.r = robjects.r
        self.report_template = report_template \
            or os.path.expanduser(self.knitr_report)
        self.r('require("knitr")')

    def clean(self, everything=False):
        log.debug('cleaning garbage')
        unwanted = ['aux', 'log', 'out', 'tex']
        if everything:
            unwanted.append('pdf')
        unwanted = map(lambda x: 'report.{}'.format(x), unwanted)
        for filename in unwanted:
            if os.path.exists(filename):
                log.debug('deleting {}'.format(filename))
                os.remove(filename)

        if os.path.exists('figure'):
            shutil.rmtree('figure')

    def process(self):
        log.info('processing report')
        self.r('knit2pdf("{}")'.format(self.report_template))



# -*- coding: utf-8 -*-
# vim:fenc=utf-8

'''
  Database utils
  --------------

  :copyright (c) 2014 Xavier Bruhiere.
  :license: %LICENCE%, see LICENSE for more details.
'''

import os
import copy

# We will use these settings later in the code to
# connect to the RethinkDB server.
DB_CONFIG = {
    'host': os.getenv('DB_HOST', 'localhost'),
    'port': os.getenv('DB_PORT', 28015),
    'db': os.getenv('DB_NAME', 'intuition'),
    'user': os.getenv('DB_USER', 'quant'),
    'password': os.getenv('DB_PASSWORD', 'ity')
}


def portfolio_to_dict(portfolio):
    json_pf = copy.deepcopy(portfolio).__dict__
    if json_pf['positions']:
        for sid, infos in portfolio.positions.iteritems():
            json_pf['positions'][sid] = infos.__dict__
    return json_pf



# -*- coding: utf-8 -*-
# vim:fenc=utf-8

'''
  Insights Forex live source
  --------------------------

  :copyright (c) 2014 Xavier Bruhiere
  :license: Apache 2.0, see LICENSE for more details.
'''


import time
import pandas as pd
import dna.logging
import intuition.data.forex as forex

log = dna.logging.logger(__name__)


class Forex(object):
    '''
    At each event datetime of the provided index, ForexLiveSource fetchs live
    forex data from TrueFX.
    '''

    def __init__(self, pairs, properties):
        self._wait_retry = properties.get('retry', 10)
        self.forex = forex.TrueFX(pairs=pairs)
        self.forex.connect()

    def get_data(self, sids):
        while True:
            rates = self.forex.query_rates()
            if len(rates.keys()) >= len(sids):
                log.debug('Data available for {}'.format(rates.keys()))
                break
            log.debug('Incomplete data ({}/{}), retrying in {}s'.format(
                len(rates.keys()), len(sids), self._wait_retry))
            time.sleep(self._wait_retry)
            debug_feedback = self.forex.connect()
            log.info('New Truefx connection: {}'.format(debug_feedback))

        return rates

    @property
    def mapping(self):
        return {
            'dt': (lambda x: x, 'dt'),
            #TODO Here conversion (weird result for now)
            # Or: (lambda x: pd.tslib.i8_to_pydt(x + '000000'), 'trade_time'),
            'trade_time': (lambda x: pd.datetime.utcfromtimestamp(
                float(x[:-3])), 'timeStamp'),
            'sid': (lambda x: x, 'sid'),
            'price': (float, 'bid'),
            'ask': (float, 'ask'),
            'high': (float, 'high'),
            'low': (float, 'low'),
            'volume': (lambda x: 10000, 'bid')
        }



# -*- coding: utf-8 -*-
# vim:fenc=utf-8

'''
  Backtest data source loading csv files
  --------------------------------------

  :copyright (c) 2014 Xavier Bruhiere
  :license: Apache 2.0, see LICENSE for more details.
'''

import os
import pytz
import pandas as pd
import intuition.errors


# NOTE Current limitations
#   - One sid per csv file
#   - Every csv file must have the same structure
#   - Every csv file must have an index column
class PanelCSVSource(object):
    '''
    doc: Loads data from given csv file(s), one sid per file.
    parameters:
      path: custom path to search csv [default ~/.intuition/data/
      index_column: the header of the column holding dates [default Date]
    '''

    def __init__(self, sids=[], properties={}):
        self.index_col = properties.get('index_column', 'Date')
        self.custom_path = properties.get('path')

    def _filepath(self, sid):
        csv_dir = self.custom_path or os.path.join(
            os.path.expanduser('~/.intuition'), 'data')
        file_path = os.path.join(csv_dir, sid)
        if file_path.rfind('csv') == -1:
            file_path += '.csv'

        if not os.path.exists(file_path):
            raise intuition.errors.LoadDataFailed(
                reason='{} not found'.format(file_path), sids=sid)

        return file_path

    def get_data(self, sids, start, end):
        tmp_data = {}
        for sid in sids:

            file_path = self._filepath(sid)

            # Load the entire file
            df = pd.read_csv(
                file_path, index_col=self.index_col, parse_dates=True
            )
            # Try to store mapping fields
            self._fields = map(
                lambda x: x.lower().replace(' ', '_'), df.columns
            )

            # Make sure of chronological order
            df = df.sort_index()
            # Remove extra dates
            df = df[df.index >= start]
            df = df[df.index <= end]
            # Make it UTC aware
            df.index = df.index.tz_localize(pytz.utc)
            tmp_data[sid] = df

        if len(sids) == 1:
            data = tmp_data[sid]
        else:
            data = pd.Panel(tmp_data)
        return data

    @property
    def mapping(self):
        # Try to dynamically read fields
        mapping_struct = {
            'dt': (lambda x: x, 'dt'),
            'sid': (lambda x: x, 'sid'),
        }
        # We need a field to be the standard price
        if 'adjusted_close' in self._fields:
            mapping_struct['price'] = \
                (float, self._fields.pop(self._fields.index('adjusted_close')))
        elif 'low' in self._fields:
            mapping_struct['price'] = \
                (float, self._fields.pop(self._fields.index('low')))
        else:
            raise NotImplementedError(
                'only support low and adjusted_close as price')

        for field in self._fields:
            mapping_struct[field] = (float, field)

        return mapping_struct



# -*- coding: utf-8 -*-
# vim:fenc=utf-8

'''
  Test framework for insights modules
  -----------------------------------

  :copyright (c) 2014 Xavier Bruhiere
  :license: Apache 2.0, see LICENSE for more details.
'''

import abc
import unittest
from nose.tools import ok_, nottest
import pytz
import pandas as pd
import datetime as dt
import random
import yaml
import zipline.protocol
from dna.test_utils import setup_logger, teardown_logger


def _generate_sid_data(sid):
    return {
        'source_id': __name__ + '-1234',
        'type': 4,
        'sid': sid,
        'dt': pd.tslib.Timestamp('2012/06/05', tz=pytz.utc),
        'datetime': pd.tslib.Timestamp('2012/06/05', tz=pytz.utc),
        'price': random.random() * 10,
        'volume': random.random() * 1000
    }


def generate_fake_returns(sids):
    dates = pd.date_range('2012/01/01', '2012/06/01', tz=pytz.utc)
    return pd.DataFrame(
        {sid: [random.random() * 100 for _ in range(len(dates))]
            for sid in sids},
        index=dates
    )


def generate_portfolio(sids):
    ''' Build a random, zipline compliant portfolio '''
    pf = zipline.protocol.Portfolio()
    if isinstance(sids, list):
        sids = {sid: random.randint(1, 100) for sid in sids}

    for sid, amount in sids.iteritems():
        pf.positions[sid] = zipline.protocol.Position(sid)
        pf.positions[sid].amount = amount
        cost = random.randrange(1, 200)
        pf.positions[sid].cost_basis = cost
        pf.positions[sid].last_sale_price = random.randrange(
            int(cost), int(cost + 1000 * random.random())
        )

    return pf


# pylint: disable=R0921
class FactoryAlgorithmTestCase(unittest.TestCase):
    '''
    New algorithm tests inherit from this factory class. The main idea is to
    provide a common ground to validate Intuition compliant algorithms, and
    ease tests writting.
    '''

    __metaclass__ = abc.ABCMeta

    def setUp(self):
        setup_logger(self)
        self.default_identity = 'johndoe'
        self.event_data = zipline.protocol.BarData()
        for sid in ['goog', 'aapl', 'msft']:
            self.event_data[sid] = zipline.protocol.SIDData(
                _generate_sid_data(sid)
            )

    def tearDown(self):
        teardown_logger(self)

    def _check_yaml_doc(self, Algo):
        doc = yaml.load(Algo.__doc__)
        ok_(isinstance(doc, dict))
        self.assertIn('doc', doc)
        ok_(isinstance(doc.get('parameters', {}), dict))

    def _check_signal_sid(self, sid_data):
        self.assertIsInstance(sid_data, zipline.protocol.SIDData)
        for info in _generate_sid_data('').keys():
            self.assertIn(info, sid_data)

    def _check_event_output(self, signals):
        if signals is not None:
            self.assertIsInstance(signals, dict)
            for signal_type in ['buy', 'sell']:
                for sid in signals.get(signal_type, {}):
                    self._check_signal_sid(signals[signal_type][sid])


# pylint: disable=R0921
class FactoryManagerTestCase(unittest.TestCase):
    '''
    This abstract factory class targets the same achievements as the
    FactoryAlgorithmTestCase, focusing on Intuition managers building block.
    '''

    __metaclass__ = abc.ABCMeta

    def setUp(self):
        setup_logger(self)
        self.some_date = dt.datetime(2014, 04, 10)
        self.buy_signal = {'goog': 34}
        self.test_sids = ['goog', 'aapl', 'msft']
        self.test_pf = generate_portfolio(self.test_sids)

    def tearDown(self):
        teardown_logger(self)

    @nottest
    def _check_initialization(self, manager):
        ok_(not manager.date)
        ok_(not manager.portfolio)
        ok_(not manager.perfs)
        ok_(hasattr(manager, 'log'))

    @nottest
    def _check_optimize_return(self, alloc, e_ret, e_risk):
        ok_(isinstance(alloc, dict))
        ok_(e_ret >= 0 and e_ret <= 1)
        ok_(e_risk >= 0 and e_risk <= 1)



# -*- coding: utf-8 -*-
# vim:fenc=utf-8

'''
  File context
  ------------

  :copyright (c) 2014 Xavier Bruhiere.
  :license: Apache 2.0, see LICENSE for more details.
'''

import intuition.utils
import intuition.constants
from intuition.api.context import ContextFactory


class FileContext(ContextFactory):
    '''
    Fetch the configuration from json or yaml flat files. It expects an input
    like : <path/to/file.{json,yml,yaml}> and supports relative and absolute
    path.
    However, if the first directory provided is 'intuition', it will
    search for the file into ~/.intuition.
    '''

    def initialize(self, storage):
        config_file = storage['path'][-1]
        if storage['path'][0] == 'intuition':
            # We got a relative path, try at the default location
            storage['path'][0] = intuition.constants.DEFAULT_HOME
        self.configfile = '/'.join(storage['path'])

        if config_file.find('json') > 0:
            self.fmt_module = __import__('json')
        elif (config_file.find('yaml') > 0) or (config_file.find('yml') > 0):
            self.fmt_module = __import__('yaml')
        else:
            raise NotImplementedError(
                'unsupported file format: %s' % config_file)

    def load(self):
        try:
            context = self.fmt_module.load(open(self.configfile, 'r'))
        except IOError, error:
            raise ValueError(
                'loading configuration file: {} ({})'.format(
                    error, self.configfile))

        return context



# -*- coding: utf-8 -*-
# vim:fenc=utf-8

'''
  Backtest data source fetching from yahoo finance
  ------------------------------------------------

  :copyright (c) 2014 Xavier Bruhiere
  :license: Apache 2.0, see LICENSE for more details.
'''

import zipline.utils.factory as factory


class YahooPrices(object):
    '''
    doc: Fetchs prices for the given sids from yahoo.com
    '''

    def __init__(self, sids, properties):
        pass

    @property
    def mapping(self):
        return {
            'dt': (lambda x: x, 'dt'),
            'sid': (lambda x: x, 'sid'),
            'price': (float, 'price'),
            'volume': (int, 'volume'),
        }

    def get_data(self, sids, start, end):
        return factory.load_from_yahoo(
            stocks=sids, indexes={}, start=start, end=end)


class YahooOHLC(object):
    '''
    doc: Fetchs OHLC data for the given sids from yahoo.com
    '''

    def __init__(self, sids, properties):
        pass

    @property
    def mapping(self):
        return {
            'dt': (lambda x: x, 'dt'),
            'sid': (lambda x: x, 'sid'),
            'price': (float, 'price'),
            'open': (float, 'open'),
            'high': (float, 'high'),
            'close': (float, 'close'),
            'low': (float, 'low'),
            'volume': (int, 'volume'),
        }

    def get_data(self, sids, start, end):
        data = factory.load_bars_from_yahoo(
            stocks=sids, indexes={}, start=start, end=end)
        self.sids = data.items
        return data



# -*- coding: utf-8 -*-
# vim:fenc=utf-8

'''
  Mail plugin
  -----------

  :copyright (c) 2014 Xavier Bruhiere
  :license: Apache 2.0, see LICENSE for more details.
'''


import os
import time
import codecs
import requests
import jinja2
import dna.logging
import intuition.data.remote as remote
import insights.analysis as analysis

log = dna.logging.logger(__name__)


def human_sid(symbol):
    sid = remote.lookup_symbol(symbol)
    if len(sid):
        human_fmt = '{} {} - {} ({})'.format(
            sid[0]['name'],
            sid[0]['typeDisp'],
            sid[0]['exchDisp'],
            sid[0]['symbol']
        )
    else:
        human_fmt = symbol

    return human_fmt


# TODO Get some inspiration: http://godoc.org/github.com/riobard/go-mailgun
class Mailgun(object):
    '''
    Send emails through mailgun api
    '''
    _api_url = 'https://api.mailgun.net/v2/{}/messages'
    _api_key = ''

    def __init__(self, friendly_name):
        self._api_key = os.environ.get('MAILGUN_API_KEY')
        domain = os.environ.get('MAILGUN_DOMAIN')
        self.from_email = '{} <me@{}>'.format(friendly_name, domain)
        self._api_url = self._api_url.format(domain)

    def send(self, targets, subject, body, attachments=None):
        attachments = attachments or []
        attachments = [
            ('attachment', open(filename, 'rb')) for filename in attachments
        ]
        if isinstance(targets, str):
            targets = [targets]
        payload = {
            'from': self.from_email,
            'to': targets,
            'subject': subject,
            'html': body
        }
        # TODO Support multiple attachments
        feedback = requests.post(
            self._api_url,
            files=attachments,
            auth=('api', self._api_key),
            data=payload
        )
        return feedback


class Report(Mailgun):
    ''' Sent mail reports about the situation '''

    _assets_dir = os.path.expanduser('~/.intuition/assets/')
    report_name = 'report.rnw'
    mail_name = 'mail-template.html'

    def __init__(self, targets):
        Mailgun.__init__(self, 'intuition.io')
        self.targets = targets
        self._last_send = time.time()

        self.generator = analysis.Stocks()
        log.info('Mail report ready', recipients=targets)

        self.template_env = jinja2.Environment(
            loader=jinja2.FileSystemLoader(self._assets_dir))

    def _render_report_template(
            self, summary, orderbook, metrics, benchmark='FCHI'):
        stocks = []
        for sid, infos in orderbook.iteritems():
            stocks.append({
                'action': 'Buy' if infos['value'] > 0 else 'Sell',
                'quantity': abs(infos['value']),
                'name': sid,
                'symbol': infos['symbol'].upper()
            })
        template = self.template_env.get_template(self.report_name + '.j2')
        return template.render(
            summary=summary,
            start_date='2013/11/01',
            benchmark=benchmark,
            orderbook=stocks,
            metrics=metrics
        )

    def _render_email_template(self, identity, orderbook):
        completion = []
        for sid, infos in orderbook.iteritems():
            completion.append({
                'action': 'Buy' if infos['value'] > 0 else 'Sell',
                'quantity': abs(infos['value']),
                'equity': sid
            })
        template = self.template_env.get_template(self.mail_name + '.j2')
        return template.render(
            strategy_name=identity,
            suggestions=completion
        )

    @property
    def is_allowed(self):
        return time.time() - self._last_send > 30.0

    def send_briefing(
            self, __class__, manager, identity, orderbook, perf_tracker):
        # TODO Portfolio summary
        if orderbook and self.is_allowed:
            self._last_send = time.time()
            identity = identity.capitalize()
            orderbook = {
                human_sid(sid): {'symbol': sid, 'value': value}
                for sid, value in orderbook.iteritems()
            }

            # TODO __class__ output is not friendly
            summary = {
                'identity': identity,
                'algorithm': __class__.__name__,
                'manager': manager.__class__.__name__ if manager else None
            }
            log.info('generating report template')
            # TODO catch errors
            report = self._render_report_template(
                summary, orderbook, perf_tracker
            )
            fd = codecs.open(
                self._assets_dir + self.report_name, 'w', 'utf-8'
            )
            fd.write(report)
            fd.close()

            self.generator.process()

            feedback = self.send(
                targets=self.targets,
                subject='{} notification'.format(identity),
                body=self._render_email_template(identity, orderbook),
                attachments=[
                    self._assets_dir + 'legal_notice.txt',
                    'report.pdf'
                ]
            )
            log.debug(feedback.json())

            self.generator.clean(everything=False)



# -*- coding: utf-8 -*-
# vim:fenc=utf-8

'''
  Linear optimization strategies for portfolio  managers
  ------------------------------------------------------

  :copyright (c) 2014 Xavier Bruhiere.
  :license: Apache 2.0, see LICENSE for more details.
'''

from intuition.api.portfolio import PortfolioFactory


class Constant(PortfolioFactory):
    '''
    doc: Allocate a constant amount to every signals, regardless of the
      portfolio.<br \> The algorithm can dynamically change the parameters and
      also advise a "scale" value to each sid.
    parameters:
      buy_amount: Number of sid units to buy [default 100]
      sell_amount: Number of sid units to sell [default All]
    '''
    def initialize(self, configuration):
        self.log.debug(configuration)

    def optimize(self, to_buy, to_sell):
        '''
        Buy sid * parameters['buy_amount'] * parameters['scale'][sid]
        Sell sid * parameters['sell_amount'] * parameters['scale'][sid]
        '''
        allocations = {}

        # Process every stock the same way
        for s in to_buy:
            quantity = self.properties.get('buy_amount', 100)
            if s in self.properties.get('scale', {}):
                quantity *= self.properties['scale'][s]
            # Allocate defined amount to buy
            allocations[s] = int(quantity)

        # NOTE You must provide sell_amount if you want to short
        for s in to_sell:
            quantity = self.properties.get(
                'sell_amount', self.portfolio.positions[s].amount)
            if s in self.properties.get('scale', {}):
                quantity *= self.properties['scale'][s]
            # Allocate defined amount to buy
            allocations[s] = -int(quantity)

        # Defaults values
        e_ret = 0
        e_risk = 1
        return allocations, e_ret, e_risk


class Fair(PortfolioFactory):
    '''
    doc: Dispatch equal weigths for buy signals and give up everything on sell
      ones.
    '''
    def optimize(self, to_buy, to_sell):
        allocations = {}
        if to_buy:
            fraction = round(1.0 / float(len(to_buy)), 2)
            for s in to_buy:
                allocations[s] = fraction
        for s in to_sell:
            allocations[s] = - self.portfolio.positions[s].amount
        return allocations, 0, 1



#
# Copyright 2013 Xavier Bruhiere
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import requests
import os
import time


def push_to_android(api_key, device_id, payload, push_type):
    ''' Push android notification through www.pushbullet.com '''
    payload['type'] = push_type
    payload['device_id'] = device_id
    return requests.post('https://api.pushbullet.com/api/pushes',
                         data=payload,
                         auth=(api_key, ''))


class AndroidPush():
    '''
    Push Android notifications
    https://www.pushbullet.com/api
    '''

    _api_url = 'https://api.pushbullet.com/api/'
    device = None
    _count = 0
    _rate_limit = 10
    _min_interval = 30
    _last_time = -30

    def __init__(self, device_name):
        try:
            self._api_key = os.environ['PUSHBULLET_API_KEY']
        except KeyError:
            self._api_key = ''
        self.device = self._device_lookup(device_name)

    def _device_lookup(self, device_name):
        ''' Request informations about <name> device '''
        device_found = {}
        response = requests.get(self._api_url + 'devices',
                                auth=(self._api_key, ''))
        if response.ok:
            for device in response.json()['devices']:
                if device_name == device['extras']['model']:
                    #TODO Check device['extras']['nickname'] as well
                    device_found = device
        return device_found

    def _detect_push_type(self, payload):
        push_type = ''
        if 'url' in payload:
            push_type = 'link'
        elif 'address' in payload:
            push_type = 'address'
        elif 'file' in payload:
            push_type = 'file'
            #TODO File push needs multi-part form encoding
            raise NotImplementedError('file payload is not yet supported')
        elif 'items' in payload:
            push_type = 'list'
        elif 'body' in payload:
            push_type = 'note'
        else:
            raise ValueError('invalid push payload')
        return push_type

    def push(self, payload, push_type=None):
        if push_type is None:
            push_type = self._detect_push_type(payload)
        return push_to_android(self._api_key,
                               self.device['id'],
                               payload,
                               push_type)

    def _watchdog(self):
        ''' Notifications are stopped if
          * The last one was too close
          * We reach a rated limit '''
        too_early = (time.time() - self._last_time < self._min_interval)
        too_much = (self._count >= self._rate_limit)
        return (False if (too_early or too_much) else True)

    def notify(self, orderbook):
        if orderbook and self._watchdog():
            # Alert user of the orders about to be processed
            # Ok... kind of fancy method
            #ords = {'-1': 'You should sell', '1': 'You should buy'}
            items = ['{} {} stocks of {}'.format(
                #ords[str(amount / abs(amount))], amount, ticker)
                ['buy', 'sell'][amount < 0], amount, ticker)
                for ticker, amount in orderbook.iteritems()
                if amount != 0]
            payload = {
                'title': 'Portfolio manager notification',
                'items': items,
            }
            req = self.push(payload)
            if req.ok:
                data = req.json()
                print('successfully notified with id {}:{} at {}'
                      .format(data['receiver_id'],
                              data['id'],
                              data['created']))
                req.close()
            self._count += 1
            self._last_time = time.time()



# -*- coding: utf-8 -*-
# vim:fenc=utf-8

'''
  Backtest data source using database
  -----------------------------------

  :copyright (c) 2014 Xavier Bruhiere
  :license: Apache 2.0, see LICENSE for more details.
'''

import pandas as pd
import copy
from insights.plugins.database.rethink import RethinkdbFinance
from intuition.data.quandl import DataQuandl


# NOTE It would be cool if this class could inherit from the one above
class RethinkdbBackedByQuandl(object):
    '''
    doc: Fetch quotes from a <a href="http://rethinkdb.com/">Rethinkdb</a>
      database. If not found, it will try to download it from <a
      href="http://quandl.com">Quandl</a>.
    parameters:
      db: Database name [default quotes],
      select: field to considere, like open, rate, ... [default All]
      save_missing: Wether save or not in database the data downloaded
        [default false]
      offline: Wether to allow or not missing data downloading [default false]
    '''

    def __init__(self, sids, properties):
        ''' Combine rethinkdb and quandl backends '''
        # Should save dataframes downloaded from quandl ?
        self._save_missing_data = properties.get('save_missing')
        # Allow downloads ?
        self._offline = properties.get('offline')
        # Only get from db or quandl this field
        select = properties.get('select')
        self._select = [select] if select else []

        self.db = RethinkdbFinance(db=properties.get('db', 'quotes'))
        self.quandl = DataQuandl()

        # TODO Check with dates
        self._missing_sids = [sid for sid in sids
                              if not self.db.available(sid)]

        # Select the right mapping
        if self._select:
            self.mapping_choice = 'whatever_dataframe'
        else:
            if properties['universe'].exchange == 'forex':
                self.mapping_choice = 'forex_panel'
            else:
                #if properties['universe'].exchange.find('paris') > 0:
                self.mapping_choice = 'stock_panel'

    def _dl_missing_data(self, data, start, end):
        missing_data = self.quandl.fetch(
            self._missing_sids, start=start, end=end)
        # FIXME Handle symbols not found
        for sid, series in missing_data.iteritems():
            if self._select:
                if data.empty:
                    data = pd.DataFrame(
                        series[self._select], index=missing_data[sid].index)
                else:
                    data[sid] = series[self._select]
            else:
                if data.empty:
                    data = pd.Panel({sid: series})
                else:
                    data[sid] = series
            if self._save_missing_data:
                self.db.save_quotes(
                    sid, series, {}, reset=True)

        return data

    def get_data(self, sids, start, end):
        data = self.db.quotes(
            sids, start=start, end=end, select=copy.copy(self._select))

        if self._missing_sids and not self._offline:
            data = self._dl_missing_data(data, start, end)

        return data.fillna(method='pad')

    @property
    def mapping(self):
        mapping = {
            'forex_panel': {
                'dt': (lambda x: x, 'dt'),
                'sid': (lambda x: x, 'sid'),
                'price': (lambda x: x, 'rate'),
                'low': (lambda x: x, 'low_(est)'),
                'high': (lambda x: x, 'high_(est)'),
                'volume': (lambda x: 100001, 'rate')
            },
            'stock_panel': {
                'dt': (lambda x: x, 'dt'),
                'sid': (lambda x: x, 'sid'),
                'price': (lambda x: x, 'adjusted_close'),
                'open': (lambda x: x, 'open'),
                'low': (lambda x: x, 'low'),
                'high': (lambda x: x, 'high'),
                'close': (lambda x: x, 'close'),
                'volume': (lambda x: x, 'volume')
            },
            'whatever_dataframe': {
                'dt': (lambda x: x, 'dt'),
                'sid': (lambda x: x, 'sid'),
                'price': (lambda x: x, 'price'),
                'volume': (lambda x: 100001, 'price')
            }
        }
        return mapping[self.mapping_choice]



# -*- coding: utf-8 -*-
# vim:fenc=utf-8

'''
  Etcd context
  ------------

  Fetch the configuration from ectd cluster.
  https://github.com/coreos/etcd
  It needs an input like :
  <1.2.3.4:port/doc_id>

  :copyright (c) 2014 Xavier Bruhiere.
  :license: Apache 2.0, see LICENSE for more details.
'''

# Can't make python-etcd to work currently
import etcd
from intuition.api.context import ContextFactory


class EtcdContext(ContextFactory):

    def initialize(self, storage):
        # storage = {{ etcd_ip }}:{{ etcd_port }}/{{ conf_id }}
        etcd_ip = storage['uri'].split(':')[0]
        etcd_port = int(storage['uri'].split(':')[1])
        self.conf_id = storage['path'][0]
        self.client = etcd.Client(host=etcd_ip, port=etcd_port)

    def load(self):
        #TODO A beautiful recursive function
        context = {}
        for item in self.client.get('/' + self.conf_id):
            key_1 = item.key.split('/')[-1]
            if item.dir:
                context[key_1] = {}
                for subitem in self.client.get(item.key):
                    context[key_1][subitem.key.split('/')[-1]] = subitem.value
            else:
                context[key_1] = item.value

        return context



# -*- coding: utf-8 -*-
# vim:fenc=utf-8

'''
  Signal Generators
  -----------------

  :copyright (c) 2014 Xavier Bruhiere
  :license: Apache 2.0, see LICENSE for more details.
'''

import collections
import zipline.transforms.ta as ta
import insights.transforms as transforms
import insights.plugins.mail as mail
from intuition.api.algorithm import TradingFactory
from insights.algorithms.utils import common_middlewares


class RSIWithMA(TradingFactory):
    '''
    doc: |
      Use The Relative Strengh Indicator to detect signals<br \>
        - RSI < 70% => sell<br \>
        - RSI > 30% => buy<br \>
      Then check Long and Short Moving averages to confirm:<br />
        - Process if MAs are getting closer'
    parameters:
      buy_trigger: Low percent of the RSI [default 30]
      sell_trigger: High percent of the RSI [default 70]
      period: RSI computation period of time [default 14]
      window: Quote history loopback to provide to the manager [default 40]
      only_full: Start when we have a full window of quotes [default false]
    '''

    over_priced = {}
    under_priced = {}

    def initialize(self, properties):
        # Interactive, mobile, hipchat, database and commission middlewares
        for middleware in common_middlewares(properties, self.identity):
            self.use(middleware)

        report_mails = properties.get('reports')
        if report_mails:
            self.use(mail.Report(report_mails).send_briefing)

        self.buy_trigger = properties.get('buy_trigger', 30)
        self.sell_trigger = properties.get('sell_trigger', 70)
        self.period = properties.get('period', 14)

        # RSI Signal
        self.rsi = ta.RSI(timeperiod=self.period)

        # Quotes loopback for managers
        self.prices_transform = transforms.get_past_prices(
            window_length=properties.get('window', 40),
            compute_only_full=properties.get('only_full')
        )

    def warm(self, data):
        self.over_priced = {sid: False for sid in data}
        self.under_priced = {sid: False for sid in data}

    def event(self, data):
        signals = {}

        rsi_data = self.rsi.handle_data(data)
        loopback_prices = self.prices_transform.handle_data(data)
        if rsi_data.isnull().any() or (loopback_prices is None):
            return signals

        self.manager.advise(historical_prices=loopback_prices)

        ranked = {}
        banked = {}
        recorded_state = {}
        for sid in data:
            rsi = rsi_data[sid]
            # NOTE portfolio:positions:<sid>:amount
            recorded_state[sid] = {
                'price': data[sid].price,
                'rsi': rsi
            }
            if rsi < self.sell_trigger and self.over_priced[sid]:
                recorded_state[sid]['msg'] = (
                    'Selling {}: RSI crossed down sell trigger ({} < {})'
                    .format(sid, rsi, self.sell_trigger)
                )
                banked[sid] = rsi
                self.over_priced[sid] = False

            elif rsi > self.sell_trigger and not self.over_priced[sid]:
                recorded_state[sid]['msg'] = (
                    '{} is now overpriced ({} > {})'
                    .format(sid, rsi, self.sell_trigger)
                )
                self.over_priced[sid] = True

            elif rsi > self.buy_trigger and self.under_priced[sid]:
                recorded_state[sid]['msg'] = (
                    'Buying {}: RSI crossed up "buy trigger" ({} > {})'
                    .format(sid, rsi, self.buy_trigger)
                )
                ranked[sid] = rsi
                self.under_priced[sid] = False

            elif rsi < self.buy_trigger and not self.under_priced[sid]:
                recorded_state[sid]['msg'] = (
                    '{} is now underpriced ({} < {})'
                    .format(sid, rsi, self.buy_trigger)
                )
                self.under_priced[sid] = True

        signals['buy'] = collections.OrderedDict(
            {sid: data[sid] for sid, _ in sorted(
                ranked.items(), key=lambda t: t[1])})
        signals['sell'] = collections.OrderedDict(
            {sid: data[sid] for sid, _ in sorted(
                banked.items(), key=lambda t: t[1])})

        self.record(**recorded_state)
        return signals



# -*- coding: utf-8 -*-
# vim:fenc=utf-8

'''
  Algorithm utilities
  -------------------

  :copyright (c) 2014 Xavier Bruhiere.
  :license: Apache 2.0, see LICENSE for more details.
'''

import insights.plugins.database.rethink as database
import insights.plugins.mobile as mobile
import insights.plugins.hipchat as hipchat
import insights.plugins.messaging as msg


def common_middlewares(properties, identity):
    middlewares = []

    if properties.get('interactive'):
        middlewares.append(msg.RedisProtocol(identity).check)

    device = properties.get('mobile')
    if device:
        middlewares.append(mobile.AndroidPush(device).notify)

    if properties.get('save'):
        middlewares.append(database.RethinkdbFinance(
            table=identity, db='portfolios', reset=True)
            .save_portfolio)

    hipchat_room = properties.get('hipchat')
    if hipchat_room:
        middlewares.append(hipchat.Bot(
            hipchat_room, name=identity).notify)

    return middlewares



# -*- coding: utf-8 -*-
# vim:fenc=utf-8

'''
  Dummy algorithms
  ----------------

  :copyright (c) 2014 Xavier Bruhiere.
  :license: Apache 2.0, see LICENSE for more details.
'''

import random
from intuition.api.algorithm import TradingFactory
import insights.transforms as transforms
from insights.algorithms.utils import common_middlewares


class BuyAndHold(TradingFactory):
    '''
    doc: Depending on the given parameters, this algorithm will buy every sids
      once or at regular intervals.
    parameters:
      start_time: Unique "buy signal".
        Ignored if equal or less than 0 [default 3]
      rate: Regular "buy signal", ignored if -1 [default -1]
    '''
    def initialize(self, properties):
        # Punctual buy signals, except 0 means never
        self.start_time = properties.get('start_time', 3)
        # regularly buy signals
        self.rate = properties.get('rate', -1)

        # Interactive, mobile, hipchat, database and commission middlewares
        for middleware in common_middlewares(properties, self.identity):
            self.use(middleware)

    def _check_rate(self):
        return (self.rate > 0) and (self.elapsed_time.days % self.rate == 0)

    def event(self, data):
        signals = {'buy': {}, 'sell': {}}

        # One shot or always buying or regularly
        if self.elapsed_time.days == self.start_time \
                or self._check_rate():
            # Only cares about buying everything
            signals['buy'] = data

        return signals


class Random(TradingFactory):
    '''
    doc: Randomly choose for each sid to buy, sell or pass
    parameters:
      buy_trigger: Chances to buy [default 0.5 < x < 1]
      sell_trigger: Chances to sell [default 0 < x < 0.5]
    '''
    def initialize(self, properties):

        for middleware in common_middlewares(properties, self.identity):
            self.use(middleware)

        self.buy_trigger = properties.get(
            'buy_trigger', random.randint(500, 1000) / 1000.0)
        self.sell_trigger = properties.get(
            'sell_trigger', random.randint(0, 500) / 1000.0)

        # Makes results reproductible
        random.seed(self.identity)

    def event(self, data):
        signals = {'buy': {}, 'sell': {}}
        # One shot or always buying or regularly
        for sid in data:
            luck = random.random()
            if luck >= self.buy_trigger:
                signals['buy'][sid] = data[sid]
            elif luck <= self.sell_trigger:
                signals['sell'][sid] = data[sid]

        return signals


# TODO Merge with buyandhold
# https://www.quantopian.com/posts/global-minimum-variance-portfolio?c=1
class RegularRebalance(TradingFactory):
    '''
    doc: Reconsidere the portfolio allocation regularly, providing to the
      portfolio strategy a moving window of quotes data.
    parameters:
      rate: Reallocation period [default 10]
      window: quotes window tracked [default 40]
    '''

    def initialize(self, properties):

        for middleware in common_middlewares(properties, self.identity):
            self.use(middleware)

        # Set Max and Min positions in security
        self.max_notional = 1000000.1
        self.min_notional = -1000000.0

        # This is the lookback window that the entire algorithm depends on
        self.refresh_period = properties.get('rate', 10)
        self.returns_transform = transforms.get_past_returns(
            refresh_period=self.refresh_period,
            window_length=properties.get('window', 40),
            compute_only_full=True)

    def event(self, data):
        signals = {'buy': {}, 'sell': {}}

        #get 20 days of prices for each security
        daily_returns = self.returns_transform.handle_data(data)

        #circuit breaker in case transform returns none
        #circuit breaker, only calculate every given rate
        if (daily_returns is None) or \
                (self.elapsed_time.days % self.refresh_period is not 0):
            return signals

        #reweight portfolio
        for sid in data:
            signals['buy'][sid] = data[sid]

        self.manager.advise(historical_returns=daily_returns)
        return signals



# -*- coding: utf-8 -*-
# vim:fenc=utf-8

'''
  Rethinkdb plugin
  ----------------

  :copyright (c) 2014 Xavier Bruhiere.
  :license: Apache 2.0, see LICENSE for more details.
'''

import pytz
import datetime as dt
import random
from clint.textui import progress
import pandas as pd
import rethinkdb as rdb
import dna.logging
import dna.time_utils
from intuition.data.quandl import DataQuandl
import intuition.data.utils as datautils
import insights.plugins.database.utils as dbutils


log = dna.logging.logger(__name__)


class RethinkdbBackend(object):
    ''' Low level of rethinkdb management '''

    def __init__(self, **kwargs):
        host = kwargs.get('host', dbutils.DB_CONFIG['host'])
        port = kwargs.get('port', dbutils.DB_CONFIG['port'])
        db = kwargs.get('db', dbutils.DB_CONFIG['db'])
        self.table = kwargs.get('table', None)

        try:
            self.session = self._connection(host, port, db)
        except rdb.RqlDriverError, error:
            log.error(error)
            raise

        # Prepare the database
        if db not in rdb.db_list().run(self.session):
            log.info('creating database {}'.format(db))
            rdb.db_create(db).run(self.session)
            self.session = self._connection(host, port, db)
        if kwargs.get('reset') and self.table:
            log.info('clearing table {}'.format(self.table))
            self._reset_data(self.table)

    def _reset_data(self, table):
        if table in rdb.table_list().run(self.session):
            result = rdb.table_drop(table).run(self.session)
            assert result['dropped'] == 1

        result = rdb.table_create(table).run(self.session)
        result = rdb.table(table).index_create('date').run(self.session)
        return result.get('created', 0) == 1

    def _connection(self, host, port, db):
        return rdb.connect(host=host, port=port, db=db)

    def available(self, table):
        # TODO Check with dates
        return datautils.clean_sid(table) in rdb.table_list().run(self.session)

    def random_tables(self, n=1000):
        log.info('generating random list of {} tables'.format(n))
        tables = rdb.table_list().run(self.session)
        random.shuffle(tables)
        return map(str, tables[:n])

    def last_chrono_entry(self, table):
        return rdb.table(datautils.clean_sid(table))\
            .order_by(rdb.desc('date'))\
            .limit(1)\
            .pluck(['date'])\
            .run(self.session)[0]

    #def __del__(self):
        #FIXME RqlClientError(u'MALFORMED PROTOBUF (missing field
        #`type`):\ntoken: 508\n1: 4\n5: 1\n')
        #self.session.close()


class RethinkdbFinance(RethinkdbBackend):
    '''
    Adds RethinkDB database backend to the portfolio
    '''

    def save_portfolio(self, datetime, portfolio, perf_tracker, recorded_vars):
        '''
        Store in Rethinkdb a zipline.Portfolio object
        '''
        if perf_tracker.progress != 0.0 and self.table:

            metrics = perf_tracker.to_dict()
            metrics.update({
                'date': datetime,
                'recorded': recorded_vars,
                'portfolio': dbutils.portfolio_to_dict(portfolio)
            })

            log.debug('Saving metrics in database')
            log.debug(
                rdb.table(self.table)
                   .insert(metrics)
                   .run(self.session)
            )

    def save_quotes(self, table, data, metadata={}, reset=False):
        table = datautils.clean_sid(table)
        if reset:
            self._reset_data(table)
        length = len(data)
        for date, row in progress.bar(data.iterrows(), expected_size=length):
            record = row.to_dict()
            record.update({'date': date})
            record.update(metadata)
            report = rdb.table(table).insert(record).run(self.session)
            assert not report['errors']

    def _load_quotes(self, sids, start, end, select):
        is_panel = not len(select) == 1
        data = {}
        for table in sids:
            if not self.available(table):
                log.warning('{} not found in database, skipping'
                            .format(table))
                continue

            if select:
                select.append('date')
                cursor_data = rdb.table(datautils.clean_sid(table))\
                    .filter(lambda row: row['date'].during(
                            start, end))\
                    .pluck(select)\
                    .run(self.session)
            else:
                # TODO pop 'id' field
                cursor_data = rdb.table(datautils.clean_sid(table))\
                    .filter(lambda row: row['date'].during(
                            start, end))\
                    .run(self.session)

            data[table] = {}
            for row in cursor_data:
                # Remove rethinkdb automatic id
                row.pop('id', None)
                # tzinfo of the object is rethinkdb specific
                date = row.pop('date').astimezone(pytz.utc)
                data[table][date] = row if is_panel else row[select[0]]

        if is_panel:
            # FIXME Missing data
            data = {k: v for k, v in data.iteritems() if len(v) > 0}
            data = pd.Panel(data).transpose(0, 2, 1)
        else:
            data = pd.DataFrame(data).fillna(method='pad')
        return data

    def quotes(self, sids, start=None, end=None, select=[]):
        #sids = map(str.lower, map(str, sids))
        if start:
            start = rdb.epoch_time(dna.time_utils.UTC_date_to_epoch(start))
        if end:
            # TODO Give a notice when given end is > database end
            end = rdb.epoch_time(dna.time_utils.UTC_date_to_epoch(end))
        else:
            end = rdb.now()
        if select == 'ohlc':
            # TODO Handle 'adjusted_close'
            select = ['open', 'high', 'low', 'close', 'volume']
        if not isinstance(select, list):
            select = [select]

        return self._load_quotes(sids, start, end, select)


class Keeper(object):
    ''' Fill and synchronize database '''

    def __init__(self, db='quotes'):
        self.db = RethinkdbFinance(db=db)

        log.info('using {} as data provider'.format('quandl.com'))
        self.feed = DataQuandl()

    def _store(self, sid, data):
        if not data.isnull().any().any():
            log.info('saving {} in database'.format(sid))
            self.db.save_quotes(sid, data, {}, reset=self.reset)
        else:
            log.warning('{}: empty dataset'.format(sid))

    def fill_database(self, sids, start, end):
        if isinstance(end, str):
            start = dna.time_utils.normalize_date_format(start)
        if isinstance(end, str):
            end = dna.time_utils.normalize_date_format(end)
        self.reset = True
        self._download_and_store(sids, start, end)

    def sync(self, sids, start, end):
        # TODO Per quote sync
        younger_date = dt.datetime(3020, 1, 1, tzinfo=pytz.utc)
        for sid in sids:
            last_entry = self.db.last_chrono_entry(sid)
            if last_entry['date'] < younger_date:
                younger_date = last_entry['date']

        start = younger_date + dt.timedelta(1)
        end = dt.date.today()
        self.reset = False

        log.info('downloading {} data ({} -> {})'.format(sids, start, end))
        self._download_and_store(sids, start, end)

    def _download_and_store(self, sids, start, end):
        data = self.feed.fetch(
            sids, start=start, end=end, returns='pandas')

        for sid in data:
            try:
                self._store(sid=sid, data=data[sid])
            except Exception as error:
                log.error(error)
                continue



#
# Copyright 2014 Xavier Bruhiere
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import redis
import json
import os


BROKER_CONFIG = {
    'host': os.getenv('BROKER_HOST', 'localhost'),
    'port': int(os.getenv('BROKER_PORT', 6379)),
}


class RedisProtocol():
    '''
    Messaging skills for algorithms
    It uses redis to listen for incoming messages and publish informations
    Goal is to provide algo <-> algo and user <-> algo communication

    redis 172.17.0.3:6379> rpush "intuition" "{'JRA': 12}"
    '''

    # How much time are we waiting for incoming messages
    timeout = 1

    def __init__(self, channel='intuition'):
        self.channel = channel
        self.client = redis.StrictRedis(
            host=BROKER_CONFIG['host'], port=BROKER_CONFIG['port'], db=0)

    def check(self, order, sids):
        ''' Check if a message is available '''
        payload = "{}"
        #TODO store hashes {'intuition': {'id1': value, 'id2': value}}
        # Block self.timeout seconds on self.channel for a message
        raw_msg = self.client.blpop(self.channel, timeout=self.timeout)
        if raw_msg:
            _, payload = raw_msg

            msg = json.loads(payload.replace("'", '"'), encoding='utf-8')
            for sid in msg.keys():
                #TODO Harmonize lower() and upper() symbols
                if sid.lower() in map(str.lower, sids):
                    print('ordering {} of {}'.format(msg[sid], sid))
                    #order(sid.upper(), msg[sid])
                    order(sid, msg[sid])
                else:
                    print('skipping unknown symbol {}'.format(sid))

    def emit(self, data):
        #NOTE Someway to control the correct execution ?
        self.client.rpush(self.channel, data)



# -*- coding: utf-8 -*-
# vim:fenc=utf-8

'''
  Packaging
  ---------

  :copyright (c) 2014 Xavier Bruhiere
  :license: Apache 2.0, see LICENSE for more details.
'''


import os
from glob import glob
from setuptools import setup, find_packages

from insights import __version__, __author__, __licence__


requires = [
    'numpy>=1.8.1',
    'patsy>=0.2.1',
    'clint>=0.3.7',
    'python-etcd>=0.3.0',
    'redis>=2.9.1',
    'rpy2>=2.3.10',
    'pymongo>=2.7.1',
    'PyYAML>=3.11',
    'requests>=2.3.0',
    'rethinkdb>=1.12.0-2',
    'scipy==0.14.0',
    'pandas==0.13.1',
    'influxdb>=0.1.7',
    'intuition>=0.4.2'
]


def long_description():
    try:
        with open('README.md') as f:
            return f.read()
    except IOError:
        return "failed to read README.md"

setup(
    name='insights',
    version=__version__,
    description=('Quantitative algorithms, portfolio managers, '
                 'data sources and contexts for Intuition'),
    author=__author__,
    author_email='xavier.bruhiere@gmail.com',
    packages=find_packages(),
    long_description=long_description(),
    license=__licence__,
    install_requires=requires,
    url="https://github.com/hackliff/insights",
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Operating System :: OS Independent',
        'Intended Audience :: Science/Research',
        'Topic :: Office/Business :: Financial',
        'Topic :: Scientific/Engineering :: Information Analysis',
        'Topic :: System :: Distributed Computing',
    ],
    scripts=['scripts/intuition-db'],
    data_files=[
        (os.path.expanduser('~/.intuition/R'), glob('./R/*')),
        (os.path.expanduser('~/.intuition/assets'), glob('./assets/*'))
    ])



# -*- coding: utf-8 -*-
# vim:fenc=utf-8

'''
  Influxdb plugin
  ---------------

  :copyright (c) 2014 Xavier Bruhiere.
  :license: Apache 2.0, see LICENSE for more details.
'''

from influxdb import client as influxdb
import dna.logging
import dna.time_utils
import insights.plugins.database.utils as dbutils


log = dna.logging.logger(__name__)


class InfluxdbBackend():
    '''
    Adds InfluxDB database backend to the portfolio
    '''
    def __init__(self, name):
        self.name = name
        self.session = self._connection()

    def _connection(self):
        #TODO No error raised if it fails, check it yourself
        return influxdb.InfluxDBClient(
            dbutils.DB_CONFIG['host'], dbutils.DB_CONFIG['port'],
            dbutils.DB_CONFIG['user'], dbutils.DB_CONFIG['password'],
            dbutils.DB_CONFIG['db'])

    def save_portfolio(self, datetime, portfolio):
        '''
        Store in Rethinkdb a zipline.Portfolio object
        '''
        log.debug('Saving portfolio in database')
        pf = dbutils.portfolio_to_dict(portfolio)
        pf.pop('positions')
        data = [{
            "name": self.name,
            "time": datetime.strftime("%Y-%m-%d %H:%M"),
            "columns": pf.keys(),
            "points": [pf.values()]}]
        # Timestamp type is not json serializable
        data[0]['points'][0][-1] = pf['start_date'].__str__()

        self.session.write_points(data)



# -*- coding: utf-8 -*-
# vim:fenc=utf-8

'''
  hipchat Bot
  -----------

  :copyright (c) 2014 Xavier Bruhiere.
  :license: Apache 2.0, see LICENSE for more details.
'''


import os
import requests
import dna.logging

log = dna.logging.logger(__name__)


class Bot(object):
    '''
    Hipchat api client that sends notifications to a specified room
    Doc: https://www.hipchat.com/docs/api
    '''

    api_key = os.environ.get('HIPCHAT_API')
    api_url = 'https://api.hipchat.com/v1'
    name = 'Intuition Bot'
    bg_color = 'green'
    intro = 'Hey guys, I detected an opportunity'

    def __init__(self, room_id, name=None, api_key=None):
        self.room_id = room_id
        if api_key:
            self.api_key = api_key
        if name:
            self.name = name

    def _test_token(self):
        ''' TODO '''
        pass

    def _api_call(self, path, data={}, http_method=requests.get):
        ''' Process an http call against the hipchat api '''
        log.info('performing api request', path=path)
        response = http_method('/'.join([self.api_url, path]),
                               params={'auth_token': self.api_key},
                               data=data)
        log.debug('{} remaining calls'.format(
            response.headers['x-ratelimit-remaining']))
        return response.json()

    def message(self, body, room_id, style='text'):
        ''' Send a message to the given room '''
        # TODO Automatically detect body format ?
        path = 'rooms/message'
        data = {
            'room_id': room_id,
            'message': body,
            'from': self.name,
            'notify': 1,
            'message_format': style,
            'color': self.bg_color
        }
        log.info('sending message to hipchat', message=body, room=room_id)
        feedback = self._api_call(path, data, requests.post)
        log.debug(feedback)
        return feedback

    def notify(self, datetime, orderbook):
        # TODO Same flood security as mobile
        if orderbook:
            body = '<strong>{} - {}</strong><ul><li>{}</li></ul>'.format(
                datetime,
                self.intro,
                '</li><li>'.join(
                    ['{}: {}'.format(sid, quantity)
                        for sid, quantity in orderbook.iteritems()])
            )
            self.message(body, self.room_id, style='html')



# -*- coding: utf-8 -*-
# vim:fenc=utf-8

'''
  Insights fake live source
  -------------------------

  :copyright (c) 2014 Xavier Bruhiere
  :license: Apache 2.0, see LICENSE for more details.
'''

import random
import pandas as pd


class Random(object):
    '''
    At each event datetime of the provided index, FakeLiveSource
    generates random prices
    '''
    def __init__(self, sids, properties):
        pass

    @property
    def mapping(self):
        return {
            'dt': (lambda x: x, 'dt'),
            'sid': (lambda x: x, 'sid'),
            'price': (float, 'price'),
            'volume': (int, 'volume'),
        }

    def _feed_random_data(self):
        return {
            'price': 100 * random.random(),
            'volume': random.randrange(1000, 10000)
        }

    def get_data(self, sids):
        data = {}
        for sid in sids:
            data[sid] = self._feed_random_data()

        return pd.DataFrame(data)



# -*- coding: utf-8 -*-
# vim:fenc=utf-8

'''
  Insights equities live source
  -----------------------------

  :copyright (c) 2014 Xavier Bruhiere
  :license: Apache 2.0, see LICENSE for more details.
'''

import intuition.data.remote as remote


class Stocks(object):
    '''
    Download live (or slightly delayed) quotes from Google Finance
    '''

    _specific_mapping = {
        'google': {
            'price': (float, 'price'),
            'change': (float, 'perc_change'),
            'volume': (lambda x: int(10001), 'price')
        },
        # NOTE short_ration also available but often returned as NaN
        'yahoo': {
            'price': (float, 'last'),
            # NOTE Often to N/A outside US
            # 'change': (float, 'change_pct'),
            # 'pe': (float, 'PE'),
            'volume': (lambda x: int(10001), 'last')
        }
    }

    def __init__(self, sids, properties):
        self._source = properties.get('source', 'yahoo')
        self._mapping = {
            'dt': (lambda x: x, 'dt'),
            'sid': (lambda x: x, 'sid'),
        }
        if self._source in self._specific_mapping.keys():
            self._mapping.update(self._specific_mapping[self._source])
        else:
            raise ValueError('{} source not available.'.format(self._source))

    @property
    def mapping(self):
        return self._mapping

    def get_data(self, sids):
        # FIXME No volume information with this method
        if self._source == 'google':
            snapshot = remote.snapshot_google(symbols=sids)
        elif self._source == 'yahoo':
            snapshot = remote.snapshot_yahoo_pandas(symbols=sids)

        if snapshot.empty:
            raise ValueError('no equities data available')
        return snapshot
